package app.alan.alan_voice

import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry.Registrar
import android.view.View
import android.view.Gravity
import android.widget.FrameLayout
import android.app.Activity
import android.view.ViewGroup
import com.alan.alansdk.Alan
import com.alan.alansdk.AlanCallback
import com.alan.alansdk.BasicSdkListener
import com.alan.alansdk.ScriptMethodCallback
import com.alan.alansdk.button.AlanButton
import com.alan.alansdk.events.EventCommand
import com.alan.alansdk.AlanState
import com.alan.alansdk.AlanConfig
import io.flutter.plugin.common.EventChannel

class AlanVoicePlugin(registrar: Registrar): MethodCallHandler {

  private val activity: Activity = registrar.activity()
  private var alanButton: AlanButton? = null
  private var callBackChannel: EventChannel = EventChannel(registrar.messenger(), "alan_voice_callback")

  init {

  }

  companion object {

    private val ARGUMENT_PROJECT_ID = "projectId"
    private val ARGUMENT_PROJECT_AUTH_JSON = "projectAuthJson"

    private val ARGUMENT_BUTTON_HORIZONTAL_ALIGN = "buttonAlign"
    private val ARGUMENT_BUTTON_TOP_MARGIN = "topMargin"
    private val ARGUMENT_BUTTON_BOTTOM_MARGIN = "bottomMargin"
    private val ARGUMENT_STT_VISIBLE = "sttVisible"

    private val ARGUMENT_METHOD_NAME = "method_name"
    private val ARGUMENT_METHOD_ARGS = "method_args"

    private val ARGUMENT_VISUALS = "visuals"
    private val ARGUMENT_TEXT = "text"
    private val ARGUMENT_COMMAND = "command"

    @JvmStatic
    fun registerWith(registrar: Registrar) {
      val channel = MethodChannel(registrar.messenger(), "alan_voice")
      channel.setMethodCallHandler(AlanVoicePlugin(registrar))
    }
  }

  override fun onMethodCall(call: MethodCall, result: Result) {
    when (call.method) {
      "getVersion" -> version(result)
      "addButton" -> addButton(result, call)
      "removeButton" -> removeButton(result, call)
      "showButton" -> showButton(result)
      "hideButton" -> hideButton(result)
      "activate" -> activate(result)
      "deactivate" -> deactivate(result)
      "callProjectApi" -> callScript(result, call)
      "setVisualState" -> setVisualState(result, call)
      "playText" -> playText(result, call)
      "playCommand" -> playCommand(result, call)
      else -> result.notImplemented()
    }
  }

  private fun setVisualState(result: Result, call: MethodCall) {
    if (alanButton != null) {
      alanButton?.sdk?.setVisuals(call.argument<String>(ARGUMENT_VISUALS))
      result.success(true)
    } else {
      result.error("No alan button", null, null)
    }
  }

  private fun playCommand(result: Result, call: MethodCall) {
    if (alanButton != null) {
      alanButton?.playCommand(call.argument<String>(ARGUMENT_COMMAND)) { _, _, _ ->
        result.success(true)
      }
    } else {
      result.error("No alan button", null, null)
    }
  }

  private fun playText(result: Result, call: MethodCall) {
    if (alanButton != null) {
      alanButton?.playText(call.argument<String>(ARGUMENT_TEXT))
      result.success(true)
    } else {
      result.error("No alan button", null, null)
    }
  }

  private fun callScript(result: Result, call: MethodCall) {
    if (alanButton != null) {
      alanButton?.sdk?.call(call.argument<String>(ARGUMENT_METHOD_NAME),
              call.argument<String>(ARGUMENT_METHOD_ARGS)
      ) { method, body, error ->
        alanButton?.post {
          result.success(listOf(method, body, error))
        }
      }
    } else {
      result.error("No alan button", null, null)
    }
  }

  private fun version(result: Result) {
    if (alanButton != null) {
      result.success(alanButton?.sdk?.version ?: "")
    } else {
      result.error("No alan button", null, null)
    }
  }

  private fun activate(result: Result) {
    alanButton?.activate()
    result.success(true)
  }

  private fun deactivate(result: Result) {
    alanButton?.deactivate()
    result.success(true)
  }

  private fun showButton(result: Result) {
    alanButton?.showButton()
    result.success(true)
  }

  private fun hideButton(result: Result) {
    alanButton?.hideButton()
    result.success(true)
  }

  private fun removeButton(result: Result, call: MethodCall) {
    if (alanButton == null) {
      val rootView = activity.findViewById(android.R.id.content) as ViewGroup
      alanButton?.getSDK()?.clearCallbacks()
      alanButton?.getSDK()?.stop()
      rootView.removeView(alanButton)
      alanButton = null
    }
  }

  private fun addButton(result: Result, call: MethodCall) {
    if (alanButton == null) {
      createButton(call,null)
    }
    val projectId = call.argument<String>(ARGUMENT_PROJECT_ID)
    val authJson = call.argument<String>(ARGUMENT_PROJECT_AUTH_JSON)

    if (projectId != null) {
      val config = AlanConfig.builder()
              .setProjectId(projectId)
              .setDataObject(authJson)
              .build()
      alanButton?.initWithConfig(config)
      subscribe(alanButton?.sdk)
      alanButton?.showButton()

      val buttonHorizontalAlignment = call.argument<Int?>(ARGUMENT_BUTTON_HORIZONTAL_ALIGN) ?: 1
      alanButton?.setButtonAlign(buttonHorizontalAlignment)
      result.success(true)
    } else {
      result.error("No objectId, please provide objectId argument", null, null)
    }
  }

  inner class AlanEventSink : EventChannel.StreamHandler {
    private var sink : EventChannel.EventSink? = null

    override fun onListen(p0: Any?, p1: EventChannel.EventSink?) {
      sink = p1
    }

    override fun onCancel(p0: Any?) {
      sink = null
    }

    fun newAlanState(state: AlanState) {
      sink?.success(listOf("button_state_changed", state.name))
    }

    fun newCommand(payload: String) {
      sink?.success(listOf("command", payload))
    }
  }

  private fun subscribe(sdk: Alan?) {
    sdk?: return
    val alanSink = AlanEventSink()
    callBackChannel.setStreamHandler(alanSink)

    sdk.registerCallback(object : AlanCallback() {
      override fun onAlanStateChanged(alanState: AlanState) {
        alanSink.newAlanState(alanState)
      }

      override fun onCommandReceived(eventCommand: EventCommand?) {
        alanSink.newCommand(eventCommand?.data?.toString() ?: "")
      }
    })
  }

  private fun createButton(call: MethodCall, result: Result?) {
    val rootView = activity.findViewById(android.R.id.content) as ViewGroup

    alanButton = AlanButton(activity, null)
    val params = FrameLayout.LayoutParams(
            FrameLayout.LayoutParams.MATCH_PARENT, FrameLayout.LayoutParams.WRAP_CONTENT)
    params.gravity = Gravity.BOTTOM or Gravity.END
    alanButton?.let {
      it.layoutParams = params
      it.visibility = View.GONE

      rootView.addView(it)
    }

    result?.success(true)
  }
}
